{-# LANGUAGE QuasiQuotes, OverloadedStrings, UnicodeSyntax,
             FlexibleInstances, TemplateHaskell #-}
{-# OPTIONS_GHC -F -pgmF frquotes #-}
{-# OPTIONS -Wall -fno-warn-missing-signatures -fno-warn-unused-imports #-}
module NaPaKit where

import Prelude hiding (words)

-- The writer monad is used to get the "do" notation
import Control.Monad.Writer

import Control.Applicative

import Data.List hiding (words)
import Data.List.Split (splitOneOf)
import Data.String
import Language.LaTeX
import qualified Language.LaTeX.Builder as B
import qualified Language.LaTeX.Builder.Internal as BI
import qualified Language.LaTeX.Builder.Math as M
import qualified Language.LaTeX.Builder.Color as C
import qualified Language.LaTeX.Length as L
import Language.LaTeX.Slicer (slice,(^$))
import Language.LaTeX.Builder.QQ
import Language.LaTeX.Unicode

[keys|paragraphName|]

-- would be useful in HLaTeX

-- maybe a type class
fromChar :: Char -> LatexItem
fromChar = fromString . pure

-- shortcuts
quad = B.math M.quad
thickspace = B.math M.thickspace
mediumspace = B.math M.mediumspace
emph = B.emph
cite = B.cite
ref = B.ref
th n = n ⊕ B.textsuperscript "th"
nth = th (B.math M.n)
mc :: Monoid a => [a] -> a
mc = mconcat
ma :: [MathItem] -> LatexItem
ma = B.math . mc
(∼) = B.nbsp
ie = B.emph «i.e.»

todo :: a -> a
todo = id
{-# DEPRECATED todo "You have something to do here" #-}

usepackage xs = BI.usepackage xs . BI.pkgName
usepackages = mc . map (usepackage [])

sigplanconf = BI.otherDocumentClassKind "sigplanconf"

-- The document class of your document
docclass = B.documentclass sigplanconf []

authorinfo (name, inst, email) =
  BI.preambleCmdArgs "authorinfo" $ map BI.mandatory [name, inst, email]

titlebanner = BI.preambleCmdArg "titlebanner"

mkabstract = BI.parEnv "abstract" [] . BI.latexItem

red = C.textcolor C.red

confinfo = [qp|
\conferenceinfo{ICFP'11,} {September 19--21, 2011, Tokyo, Japan.}
\CopyrightYear{2011}
\copyrightdata{978-1-4503-0865-6/11/09}
|]

classifiers = [tex|
\category{D.3.3}{Programming Languages}{Language Constructs and Features}
                [Data types and structures; Polymorphism]
\terms{Design, Languages, Theory}
\keywords{names, binders, meta-programming, name abstraction, de Bruijn indices}
|]

document titl authors abstrct bdy = do
  let
    doc = B.document docclass preamble body
    body = B.maketitle
         ⊕ mkabstract abstrct
         ⊕ B.para classifiers
         ⊕ bdy
         ⊕ B.para [tex|\bibliographystyle{plain}|]
         ⊕ B.bibliography "bib,local"
    preamble = B.title titl
             ⊕ mc (map authorinfo authors)
             ⊕ confinfo
             ⊕ fonts
  return doc
vcenter x = B.vfill ⊕ x ⊕ B.vfill

verb mayAlign = B.texttt . B.spaceProtector (myXchar mayAlign (M.mchar B.ttchar))

-- takes the highest character to vertically align them all.
myStrut = B.vphantom (BI.rawTex "$\\{$")

myNewline = myStrut ⊕ B.decl (B.nopagebreak 4) (BI.latexParMode (B.newline ø))

width0_5 = "ᵣ₁₂ᴺᵂˢ"
width1 = "∀αβ·ƛ_⟦⟧⟨⟩↑øℓ∈"
width2 = "¬ℕ⟶⊎∨∧⊆"

align k = B.makebox (L.ex (k * 1.22)) B.centered

-- let's customize the rendering of some charcters in `verb' mode
myXchar _ _     '\n' = myNewline
myXchar mayAlign xchar x
  | x `elem` width0_5= mayAlign (0.5 :: Rational) $ xchar x
  | x `elem` width1  = mayAlign 1 $ xchar x
  | x `elem` width2  = mayAlign 2 $ xchar x
  | x `elem` "‽"     = mayAlign 0.5 " "
  | x `elem` "="     = B.ttchar x
  | x `elem` "{}"    = M.mchar B.hchar x
  | x `elem` "("     = mayAlign 1 . B.math $ M.lparen -- mleft '('
  | x `elem` ")"     = mayAlign 1 . B.math $ M.rparen -- mright ')'
myXchar _ xchar x      = xchar x

-- do not cut on unbreakable spaces
words = filter (not.null) . splitOneOf " \t\n"

-- word breakable code
wCode = B.unwords . map (verb (const id)) . words

para :: LatexItem -> ParItem
para x = B.para x -- ⊕ B.paragraph ø

namedPara :: String -> LatexItem -> ParItem
namedPara nam = BI.parNote paragraphName (BI.stringNote nam) . para

coloredNamedPara :: String -> C.Color -> C.Color -> LatexItem -> ParItem
coloredNamedPara nam bgcolor fcolor
  = (B.para « » ⊕) . B.para . (B.noindent ⊕)
  . C.fcolorbox fcolor bgcolor . B.minipage L.linewidth
  . namedPara (nam ++ " [NEEDS REVIEW]")

-- unbreakable & aligned code verbatim plain code
pCode x = put . (B.comment (filter (/='‽') x) ⊕) . B.para . ([tex|~\\~|] ⊕)
              . verb align . dropWhile (=='\n') $ x

-- λ-calculus
λCode     = wCode

-- Agda
agdaCode  = wCode
agdaPCode = pCode

-- Maths
mathW     = B.unwords . map (B.math . fromString) . words

-- Quasi Quotation

lc    = mkQQ "lc" 'λCode
agda  = mkQQ "agda" 'agdaCode
agdaP = mkQQ "agdaP" 'agdaPCode
texP  = mkQQ "texP" 'rawTexP
m     = mkQQ "m"    'mathW

-- Monadic style

put :: ParItem -> ParItemW
put = tell
pagebreak = put . B.parDecl . B.pagebreak
q :: LatexItem -> ParItemW
q = put . para
p, pr, pt :: String -> LatexItem -> ParItemW
p nam = put . namedPara nam
pr = p
pt = p
-- pr nam = put . coloredNamedPara nam (C.rgb 1 1 1) C.orange
-- pt nam = put . coloredNamedPara nam C.white C.red
mP = put . B.displaymath
section titl lbl = put . B.section $ titl ⊕ B.label lbl
subsection titl = put $ B.subsection titl
subsubsection = put . B.subsubsection
subsectionLbl    titl lbl = subsection $ titl ⊕ B.label lbl
subsubsectionLbl titl lbl = subsubsection $ titl ⊕ B.label lbl
paragraph = put . B.paragraph
figure s caption key b = B.figure s ø !$? (b >> put (B.para . B.caption $ caption ⊕ B.label key))
newpage = put B.newpage
rawTexP = put . B.para . BI.rawTex

-- vocabulary
dbis = «de Bruijn indices»
_Dbis = «De Bruijn indices»

textit = B.textit
napa = B.textsc "NaPa"
_Nameless_Painless = «{B.textbf«Na»}meless, {B.textbf«Pa»}inless»
_Agda = B.textsc «Agda»
_Template_Agda = B.textsc «Template Agda»
_PTS = ma[M.mathcal $ mc[M._P, M._T, M._S]]

-- Agda vocabulary
_ℛ1 = agdaCode "ℛ₁"
_ℛ2 = agdaCode "ℛ₂"
_A = agdaCode "A"
addN = agdaCode "addᴺ"
__ = agdaCode "_"
_ℕ = agdaCode "ℕ"
_ℛ = agdaCode "ℛ"
(⊤) = agdaCode "⊤"
(⊥) = agdaCode "⊥"
(∨) = agdaCode "∨"
ƛ = agdaCode "ƛ"
_App = agdaCode "·"
_B = agdaCode "B"
bare = agdaCode "bare"
because = agdaCode "_⟨-because_-⟩"
_Bool = agdaCode "Bool"
_BOOL = agdaCode "BOOL"
cmpN = agdaCode "cmpᴺ"
coerceN = agdaCode "coerceᴺ"
_DataKit = agdaCode "DataKit"
_Either = agdaCode "⊎"
_El = agdaCode "El"
eqN = agdaCode "==ᴺ"
exportName = agdaCode "exportName"
false = agdaCode "false"
_Fin = agdaCode "Fin"
shiftName = agdaCode "shiftName"
incl = agdaCode "⊆"
inj1 = agdaCode "inj₁"
inj2 = agdaCode "inj₂"
_Injection = agdaCode "Injection"
_Injective = agdaCode "Injective"
is42 = agdaCode "is42?"
le = agdaCode "<="
_Let = agdaCode "Let"
_Maybe = agdaCode "Maybe"
_Name0 = agdaCode "¬Nameø"
_Name = agdaCode "Name"
_Nm = agdaCode "Nm"
_ø = agdaCode "ø"
_P1 = agdaCode "_+1"
_P1Rel = agdaCode "_⟦+1⟧"
_P = agdaCode "+ᵂ"
plus = agdaCode "+"
_S1 = agdaCode "_↑1"
_S1Rel = agdaCode "_⟦↑1⟧"
_S = agdaCode "↑"
_Set0 = agdaCode "Set₀"
_Set1 = agdaCode "Set₁"
_Set = agdaCode "Set"
_Setoid = agdaCode "Setoid"
_N = agdaCode "_ᴺ"
subtractN = agdaCode "subtractᴺ"
suc = agdaCode "suc"
sucN = agdaCode "sucᴺ"
_Tm = agdaCode "Tm"
true = agdaCode "true"
_U = agdaCode "U"
_V = agdaCode "V"
_World = agdaCode "World"
zeroN = agdaCode "zeroᴺ"
zero = agdaCode "zero"

-- shortcuts

_ℛτ = _ℛ ⊕ τ
rel x = agdaCode "⟦" ⊕ x ⊕ agdaCode "⟧"
op x = agdaCode "_" ⊕ x ⊕ agdaCode "_"

completeMat :: a -> [[a]] -> [[a]]
completeMat dflt rows =
  map (take (maximum (map length rows)) . (++repeat dflt)) rows

tikzpicture :: ParItem -> ParItem
tikzpicture = BI.parEnv "tikzpicture" [] . BI.parItem

data PosLR = L | R
type Point = (Int,Int)
data Path = Path String Point PosLR LatexItem Point
type Mat = [[LatexItem]]
path :: String -> Point -> PosLR -> LatexItem -> Point -> Path
path = Path
diagram :: Mat -> [Path] -> ParItemW
diagram mat paths =
  let
    ppOptions :: [LatexItem] -> [Arg LatexItem]
    ppOptions xs | null xs   = []
                 | otherwise = [BI.optional (mconcat $ intersperse [tex|,|] xs)]
    parenArg = BI.rawArg . ('(':) . (++")")
    -- no at(...) support
    nodeOp :: [LatexItem] -> Maybe String -> LatexItem -> [Arg LatexItem]
    nodeOp options optName txt =
      mc[[BI.rawArg "node"],ppOptions options, maybe ø (pure . parenArg) optName,
         [BI.mandatory txt]]
    onPath (Path style src pos lbl dst) =
          BI.latexCmdArgs "path"
                         ([BI.optional (BI.rawTex style)
                          ,pointArg src
                          ,BI.rawArg " edge"] ⊕
                          nodeOp [onPos pos] ø lbl ⊕
                          [pointArg dst
                          ,BI.rawArg ";"])
    onPos L = [tex|left|]
    onPos R = [tex|right|]
    pointArg (x,y) = BI.rawArg (mc["(m-",show x,"-",show y,")"])
    onRow row = mc (intersperse [tex| & |] row) ⊕ [tex| \\ |]
  in
  put . tikzpicture . B.para $ [tex|\matrix (m) [matrix of math nodes, row sep=1em, column sep=1em, text height=1.5ex, text depth=0.25ex] { |]
     ⊕ mc (map onRow mat)
     ⊕ [tex| };|]
     ⊕ mc (map onPath paths)

unicodesymbols = [qp|
\providecommand\mathbbm{\mathbb}
\DeclareUnicodeCharacter{8718}{\ensuremath{\scriptstyle{\blacksquare}}}
\DeclareUnicodeCharacter{8759}{::}
\DeclareUnicodeCharacter{10214}{\ensuremath{\llbracket}} % ⟦
\DeclareUnicodeCharacter{10215}{\ensuremath{\rrbracket}} % ⟧
\DeclareUnicodeCharacter{10218}{\ensuremath{\langle\!\langle}} % ⟪
\DeclareUnicodeCharacter{10219}{\ensuremath{\rangle\!\rangle}} % ⟫
\DeclareUnicodeCharacter{7482}{\textsuperscript{N}}             % ᴺ
\DeclareUnicodeCharacter{7490}{\textsuperscript{W}}             % ᵂ
\DeclareUnicodeCharacter{411}{\textcrlambda}                    % ƛ
\DeclareUnicodeCharacter{738}{\textsuperscript{s}}              % ˢ
\DeclareUnicodeCharacter{955}{\ensuremath{\lambda}}             % λ
\DeclareUnicodeCharacter{183}{\ensuremath{\cdot}}               % ·
\DeclareUnicodeCharacter{7523}{\ensuremath{_{r}}}               % ᵣ
\DeclareUnicodeCharacter{8760}{\texttt{-}}                      % ∸ -}
\DeclareUnicodeCharacter{9679}{\ensuremath{\bullet}}            % ●
|]

fonts = usepackages ["amsfonts","amssymb","tikz","stmaryrd","url"]
      ⊕ usepackage ["pdftex","bookmarks=true","bookmarksopen=true","colorlinks=true","linkcolor=blue","citecolor=blue","urlcolor=blue"] "hyperref"
      ⊕ usepackage ["T3","OT1"] "fontenc"
      -- These two declarations are from the tipa package.
      -- However importing the packages breaks the document.
      -- Note that we require T3 fonts here.
      ⊕ [qp|\DeclareTextSymbol\textcrlambda{T3}{172}|]
      ⊕ [qp|\DeclareTextSymbolDefault\textcrlambda{T3}|]
      ⊕ [qp|\usetikzlibrary{matrix,arrows}|]
      ⊕ usepackage ["utf8x"] "inputenc"
      ⊕ unicodesymbols
